@@ -17,20 +17,15 @@ limitations under the License.
 package bigtable
 
 import (
-	"flag"
 	"fmt"
 	"math/rand"
 	"reflect"
-	"sort"
 	"strings"
 	"sync"
 	"testing"
 	"time"
 
-	"cloud.google.com/go/bigtable/bttest"
 	"golang.org/x/net/context"
-	"google.golang.org/api/option"
-	"google.golang.org/grpc"
 )
 
 func TestPrefix(t *testing.T) {
@@ -58,10 +53,6 @@ func TestPrefix(t *testing.T) {
 	}
 }
 
-var useProd = flag.String("use_prod", "", `if set to "proj,instance,table", run integration test against production`)
-var adminEndpoint = flag.String("admin-endpoint", "", "Override the admin api endpoint")
-var dataEndpoint = flag.String("data-endpoint", "", "Override the data api endpoint")
-
 func TestClientIntegration(t *testing.T) {
 	start := time.Now()
 	lastCheckpoint := start
@@ -71,52 +62,36 @@ func TestClientIntegration(t *testing.T) {
 		lastCheckpoint = n
 	}
 
-	proj, instance, table := "proj", "instance", "mytable"
-	var clientOpts []option.ClientOption
-	timeout := 20 * time.Second
-	if *useProd == "" {
-		srv, err := bttest.NewServer("127.0.0.1:0")
-		if err != nil {
-			t.Fatal(err)
-		}
-		defer srv.Close()
-		t.Logf("bttest.Server running on %s", srv.Addr)
-		conn, err := grpc.Dial(srv.Addr, grpc.WithInsecure())
-		if err != nil {
-			t.Fatalf("grpc.Dial: %v", err)
-		}
-		clientOpts = []option.ClientOption{option.WithGRPCConn(conn)}
-	} else {
-		t.Logf("Running test against production")
-		a := strings.SplitN(*useProd, ",", 3)
-		proj, instance, table = a[0], a[1], a[2]
-		timeout = 5 * time.Minute
+	testEnv, err := NewIntegrationEnv()
+	if err != nil {
+		t.Fatalf("IntegrationEnv: %v", err)
 	}
 
+	timeout := 30 * time.Second
+	if testEnv.Config().UseProd {
+		timeout = 5 * time.Minute
+		t.Logf("Running test against production")
+	} else {
+		t.Logf("bttest.Server running on %s", testEnv.Config().AdminEndpoint)
+	}
 	ctx, _ := context.WithTimeout(context.Background(), timeout)
 
-	dataClientOpts := append([]option.ClientOption(nil), clientOpts...)
-	if *dataEndpoint != "" {
-		dataClientOpts = append(dataClientOpts, option.WithEndpoint(*dataEndpoint))
-	}
-	client, err := NewClient(ctx, proj, instance, dataClientOpts...)
+	client, err := testEnv.NewClient()
 	if err != nil {
-		t.Fatalf("NewClient: %v", err)
+		t.Fatalf("Client: %v", err)
 	}
 	defer client.Close()
 	checkpoint("dialed Client")
 
-	adminClientOpts := append([]option.ClientOption(nil), clientOpts...)
-	if *adminEndpoint != "" {
-		adminClientOpts = append(adminClientOpts, option.WithEndpoint(*adminEndpoint))
-	}
-	adminClient, err := NewAdminClient(ctx, proj, instance, adminClientOpts...)
+	adminClient, err := testEnv.NewAdminClient()
 	if err != nil {
-		t.Fatalf("NewAdminClient: %v", err)
+		t.Fatalf("AdminClient: %v", err)
 	}
 	defer adminClient.Close()
 	checkpoint("dialed AdminClient")
 
+	table := testEnv.Config().Table
+
 	// Delete the table at the end of the test.
 	// Do this even before creating the table so that if this is running
 	// against production and CreateTable fails there's a chance of cleaning it up.
@@ -181,9 +156,6 @@ func TestClientIntegration(t *testing.T) {
 			{Row: "jadams", Column: "follows:tjefferson", Value: []byte("1")},
 		},
 	}
-	for _, ris := range row {
-		sort.Sort(byColumn(ris))
-	}
 	if !reflect.DeepEqual(row, wantRow) {
 		t.Errorf("Read row mismatch.\n got %#v\nwant %#v", row, wantRow)
 	}
@@ -197,7 +169,7 @@ func TestClientIntegration(t *testing.T) {
 		limit  ReadOption // may be nil
 
 		// We do the read, grab all the cells, turn them into "<row>-<col>-<val>",
-		// sort that list, and join with a comma.
+		// and join with a comma.
 		want string
 	}{
 		{
@@ -274,6 +246,79 @@ func TestClientIntegration(t *testing.T) {
 			limit:  LimitRows(2),
 			want:   "gwashington-jadams-1,jadams-tjefferson-1",
 		},
+		{
+			desc:   "read all, strip values",
+			rr:     RowRange{},
+			filter: StripValueFilter(),
+			want:   "gwashington-jadams-,jadams-gwashington-,jadams-tjefferson-,tjefferson-gwashington-,tjefferson-jadams-,tjefferson-wmckinley-,wmckinley-tjefferson-",
+		},
+		{
+			desc:   "read with ColumnFilter + row limit + strip values",
+			rr:     RowRange{},
+			filter: ChainFilters(ColumnFilter(".*j.*"), StripValueFilter()), // matches "jadams" and "tjefferson"
+			limit:  LimitRows(2),
+			want:   "gwashington-jadams-,jadams-tjefferson-",
+		},
+		{
+			desc:   "read with condition, strip values on true",
+			rr:     RowRange{},
+			filter: ConditionFilter(ColumnFilter(".*j.*"), StripValueFilter(), nil),
+			want:   "gwashington-jadams-,jadams-gwashington-,jadams-tjefferson-,tjefferson-gwashington-,tjefferson-jadams-,tjefferson-wmckinley-,wmckinley-tjefferson-",
+		},
+		{
+			desc:   "read with condition, strip values on false",
+			rr:     RowRange{},
+			filter: ConditionFilter(ColumnFilter(".*xxx.*"), nil, StripValueFilter()),
+			want:   "gwashington-jadams-,jadams-gwashington-,jadams-tjefferson-,tjefferson-gwashington-,tjefferson-jadams-,tjefferson-wmckinley-,wmckinley-tjefferson-",
+		},
+		{
+			desc:   "read with ValueRangeFilter + row limit",
+			rr:     RowRange{},
+			filter: ValueRangeFilter([]byte("1"), []byte("5")), // matches our value of "1"
+			limit:  LimitRows(2),
+			want:   "gwashington-jadams-1,jadams-gwashington-1,jadams-tjefferson-1",
+		},
+		{
+			desc:   "read with ValueRangeFilter, no match on exclusive end",
+			rr:     RowRange{},
+			filter: ValueRangeFilter([]byte("0"), []byte("1")), // no match
+			want:   "",
+		},
+		{
+			desc:   "read with ValueRangeFilter, no matches",
+			rr:     RowRange{},
+			filter: ValueRangeFilter([]byte("3"), []byte("5")), // matches nothing
+			want:   "",
+		},
+		{
+			desc:   "read with InterleaveFilter, no matches on all filters",
+			rr:     RowRange{},
+			filter: InterleaveFilters(ColumnFilter(".*x.*"), ColumnFilter(".*z.*")),
+			want:   "",
+		},
+		{
+			desc:   "read with InterleaveFilter, no duplicate cells",
+			rr:     RowRange{},
+			filter: InterleaveFilters(ColumnFilter(".*g.*"), ColumnFilter(".*j.*")),
+			want:   "gwashington-jadams-1,jadams-gwashington-1,jadams-tjefferson-1,tjefferson-gwashington-1,tjefferson-jadams-1,wmckinley-tjefferson-1",
+		},
+		{
+			desc:   "read with InterleaveFilter, with duplicate cells",
+			rr:     RowRange{},
+			filter: InterleaveFilters(ColumnFilter(".*g.*"), ColumnFilter(".*g.*")),
+			want:   "jadams-gwashington-1,jadams-gwashington-1,tjefferson-gwashington-1,tjefferson-gwashington-1",
+		},
+		{
+			desc: "read with a RowRangeList and no filter",
+			rr:   RowRangeList{NewRange("gargamel", "hubbard"), InfiniteRange("wmckinley")},
+			want: "gwashington-jadams-1,wmckinley-tjefferson-1",
+		},
+		{
+			desc:   "chain that excludes rows and matches nothing, in a condition",
+			rr:     RowRange{},
+			filter: ConditionFilter(ChainFilters(ColumnFilter(".*j.*"), ColumnFilter(".*mckinley.*")), StripValueFilter(), nil),
+			want:   "",
+		},
 	}
 	for _, tc := range readTests {
 		var opts []ReadOption
@@ -296,7 +341,6 @@ func TestClientIntegration(t *testing.T) {
 			t.Errorf("%s: %v", tc.desc, err)
 			continue
 		}
-		sort.Strings(elt)
 		if got := strings.Join(elt, ","); got != tc.want {
 			t.Errorf("%s: wrong reads.\n got %q\nwant %q", tc.desc, got, tc.want)
 		}
@@ -317,7 +361,6 @@ func TestClientIntegration(t *testing.T) {
 		t.Errorf("read RowList: %v", err)
 	}
 
-	sort.Strings(elt)
 	if got := strings.Join(elt, ","); got != want {
 		t.Errorf("bulk read: wrong reads.\n got %q\nwant %q", got, want)
 	}
@@ -447,8 +490,8 @@ func TestClientIntegration(t *testing.T) {
 	if !reflect.DeepEqual(r, wantRow) {
 		t.Errorf("Cell with multiple versions and LatestNFilter(2),\n got %v\nwant %v", r, wantRow)
 	}
-	// Check timestamp range filtering
-	r, err = tbl.ReadRow(ctx, "testrow", RowFilter(TimestampRangeFilterMicros(1000, 3000)))
+	// Check timestamp range filtering (with truncation)
+	r, err = tbl.ReadRow(ctx, "testrow", RowFilter(TimestampRangeFilterMicros(1001, 3000)))
 	if err != nil {
 		t.Fatalf("Reading row: %v", err)
 	}
@@ -471,10 +514,41 @@ func TestClientIntegration(t *testing.T) {
 	if !reflect.DeepEqual(r, wantRow) {
 		t.Errorf("Cell with multiple versions and TimestampRangeFilter(1000, 0),\n got %v\nwant %v", r, wantRow)
 	}
+	// Delete non-existing cells, no such column family in this row
+	// Should not delete anything
+	if err := adminClient.CreateColumnFamily(ctx, table, "non-existing"); err != nil {
+		t.Fatalf("Creating column family: %v", err)
+	}
+	mut = NewMutation()
+	mut.DeleteTimestampRange("non-existing", "col", 2000, 3000) // half-open interval
+	if err := tbl.Apply(ctx, "testrow", mut); err != nil {
+		t.Fatalf("Mutating row: %v", err)
+	}
+	r, err = tbl.ReadRow(ctx, "testrow", RowFilter(LatestNFilter(3)))
+	if err != nil {
+		t.Fatalf("Reading row: %v", err)
+	}
+	if !reflect.DeepEqual(r, wantRow) {
+		t.Errorf("Cell was deleted unexpectly,\n got %v\nwant %v", r, wantRow)
+	}
+	// Delete non-existing cells, no such column in this column family
+	// Should not delete anything
+	mut = NewMutation()
+	mut.DeleteTimestampRange("ts", "non-existing", 2000, 3000) // half-open interval
+	if err := tbl.Apply(ctx, "testrow", mut); err != nil {
+		t.Fatalf("Mutating row: %v", err)
+	}
+	r, err = tbl.ReadRow(ctx, "testrow", RowFilter(LatestNFilter(3)))
+	if err != nil {
+		t.Fatalf("Reading row: %v", err)
+	}
+	if !reflect.DeepEqual(r, wantRow) {
+		t.Errorf("Cell was deleted unexpectly,\n got %v\nwant %v", r, wantRow)
+	}
 	// Delete the cell with timestamp 2000 and repeat the last read,
 	// checking that we get ts 3000 and ts 1000.
 	mut = NewMutation()
-	mut.DeleteTimestampRange("ts", "col", 2000, 3000) // half-open interval
+	mut.DeleteTimestampRange("ts", "col", 2001, 3000) // half-open interval
 	if err := tbl.Apply(ctx, "testrow", mut); err != nil {
 		t.Fatalf("Mutating row: %v", err)
 	}
@@ -491,7 +565,7 @@ func TestClientIntegration(t *testing.T) {
 	}
 	checkpoint("tested multiple versions in a cell")
 
-	// Check DeleteColumnFamily
+	// Check DeleteCellsInFamily
 	if err := adminClient.CreateColumnFamily(ctx, table, "status"); err != nil {
 		t.Fatalf("Creating column family: %v", err)
 	}
@@ -535,12 +609,85 @@ func TestClientIntegration(t *testing.T) {
 			{Row: "row2", Column: "ts:col", Timestamp: 0, Value: []byte("3")},
 		},
 		"status": []ReadItem{
-			{Row: "row2", Column: "status:start", Timestamp: 0, Value: []byte("1")},
 			{Row: "row2", Column: "status:end", Timestamp: 0, Value: []byte("2")},
+			{Row: "row2", Column: "status:start", Timestamp: 0, Value: []byte("1")},
 		},
 	}
+	if !reflect.DeepEqual(r, wantRow) {
+		t.Errorf("Column family was deleted unexpectly.\n got %v\n want %v", r, wantRow)
+	}
 	checkpoint("tested family delete")
 
+	// Check DeleteCellsInColumn
+	mut = NewMutation()
+	mut.Set("status", "start", 0, []byte("1"))
+	mut.Set("status", "middle", 0, []byte("2"))
+	mut.Set("status", "end", 0, []byte("3"))
+	if err := tbl.Apply(ctx, "row3", mut); err != nil {
+		t.Errorf("Mutating row: %v", err)
+	}
+	mut = NewMutation()
+	mut.DeleteCellsInColumn("status", "middle")
+	if err := tbl.Apply(ctx, "row3", mut); err != nil {
+		t.Errorf("Delete column: %v", err)
+	}
+	r, err = tbl.ReadRow(ctx, "row3")
+	if err != nil {
+		t.Fatalf("Reading row: %v", err)
+	}
+	wantRow = Row{
+		"status": []ReadItem{
+			{Row: "row3", Column: "status:end", Timestamp: 0, Value: []byte("3")},
+			{Row: "row3", Column: "status:start", Timestamp: 0, Value: []byte("1")},
+		},
+	}
+	if !reflect.DeepEqual(r, wantRow) {
+		t.Errorf("Column was not deleted.\n got %v\n want %v", r, wantRow)
+	}
+	mut = NewMutation()
+	mut.DeleteCellsInColumn("status", "start")
+	if err := tbl.Apply(ctx, "row3", mut); err != nil {
+		t.Errorf("Delete column: %v", err)
+	}
+	r, err = tbl.ReadRow(ctx, "row3")
+	if err != nil {
+		t.Fatalf("Reading row: %v", err)
+	}
+	wantRow = Row{
+		"status": []ReadItem{
+			{Row: "row3", Column: "status:end", Timestamp: 0, Value: []byte("3")},
+		},
+	}
+	if !reflect.DeepEqual(r, wantRow) {
+		t.Errorf("Column was not deleted.\n got %v\n want %v", r, wantRow)
+	}
+	mut = NewMutation()
+	mut.DeleteCellsInColumn("status", "end")
+	if err := tbl.Apply(ctx, "row3", mut); err != nil {
+		t.Errorf("Delete column: %v", err)
+	}
+	r, err = tbl.ReadRow(ctx, "row3")
+	if err != nil {
+		t.Fatalf("Reading row: %v", err)
+	}
+	if len(r) != 0 {
+		t.Errorf("Delete column: got %v, want empty row", r)
+	}
+	// Add same cell after delete
+	mut = NewMutation()
+	mut.Set("status", "end", 0, []byte("3"))
+	if err := tbl.Apply(ctx, "row3", mut); err != nil {
+		t.Errorf("Mutating row: %v", err)
+	}
+	r, err = tbl.ReadRow(ctx, "row3")
+	if err != nil {
+		t.Fatalf("Reading row: %v", err)
+	}
+	if !reflect.DeepEqual(r, wantRow) {
+		t.Errorf("Column was not deleted correctly.\n got %v\n want %v", r, wantRow)
+	}
+	checkpoint("tested column delete")
+
 	// Do highly concurrent reads/writes.
 	// TODO(dsymonds): Raise this to 1000 when https://github.com/grpc/grpc-go/issues/205 is resolved.
 	const maxConcurrency = 100
@@ -668,9 +815,6 @@ func TestClientIntegration(t *testing.T) {
 		if err != nil {
 			t.Fatalf("Reading a bulk row: %v", err)
 		}
-		for _, ris := range row {
-			sort.Sort(byColumn(ris))
-		}
 		var wantItems []ReadItem
 		for _, val := range ss {
 			wantItems = append(wantItems, ReadItem{Row: rowKey, Column: "bulk:" + val, Value: []byte("1")})
@@ -712,12 +856,6 @@ func fill(b, sub []byte) {
 	}
 }
 
-type byColumn []ReadItem
-
-func (b byColumn) Len() int           { return len(b) }
-func (b byColumn) Swap(i, j int)      { b[i], b[j] = b[j], b[i] }
-func (b byColumn) Less(i, j int) bool { return b[i].Column < b[j].Column }
-
 func clearTimestamps(r Row) {
 	for _, ris := range r {
 		for i := range ris {
