@@ -41,6 +41,8 @@ import (
 	"sync"
 	"time"
 
+	"bytes"
+
 	emptypb "github.com/golang/protobuf/ptypes/empty"
 	"github.com/golang/protobuf/ptypes/wrappers"
 	"golang.org/x/net/context"
@@ -117,7 +119,7 @@ func (s *server) CreateTable(ctx context.Context, req *btapb.CreateTableRequest)
 	s.mu.Lock()
 	if _, ok := s.tables[tbl]; ok {
 		s.mu.Unlock()
-		return nil, fmt.Errorf("table %q already exists", tbl)
+		return nil, grpc.Errorf(codes.AlreadyExists, "table %q already exists", tbl)
 	}
 	s.tables[tbl] = newTable(req)
 	s.mu.Unlock()
@@ -182,12 +184,14 @@ func (s *server) ModifyColumnFamilies(ctx context.Context, req *btapb.ModifyColu
 	for _, mod := range req.Modifications {
 		if create := mod.GetCreate(); create != nil {
 			if _, ok := tbl.families[mod.Id]; ok {
-				return nil, fmt.Errorf("family %q already exists", mod.Id)
+				return nil, grpc.Errorf(codes.AlreadyExists, "family %q already exists", mod.Id)
 			}
 			newcf := &columnFamily{
 				name:   req.Name + "/columnFamilies/" + mod.Id,
+				order:  tbl.counter,
 				gcRule: create.GcRule,
 			}
+			tbl.counter++
 			tbl.families[mod.Id] = newcf
 		} else if mod.GetDrop() {
 			if _, ok := tbl.families[mod.Id]; !ok {
@@ -225,6 +229,7 @@ func (s *server) DropRowRange(ctx context.Context, req *btapb.DropRowRangeReques
 
 	if req.GetDeleteAllDataFromTable() {
 		tbl.rows = nil
+		tbl.rowIndex = make(map[string]*row)
 	} else {
 		// Delete rows by prefix
 		prefixBytes := req.GetRowKeyPrefix()
@@ -237,6 +242,10 @@ func (s *server) DropRowRange(ctx context.Context, req *btapb.DropRowRangeReques
 		end := 0
 		for i, row := range tbl.rows {
 			match := strings.HasPrefix(row.key, prefix)
+			if match {
+				// Delete the mapping. Row will be deleted from sorted range below.
+				delete(tbl.rowIndex, row.key)
+			}
 			if match && start == -1 {
 				start = i
 			} else if !match && start != -1 {
@@ -352,21 +361,23 @@ func streamRow(stream btpb.Bigtable_ReadRowsServer, r *row, f *btpb.RowFilter) (
 	}
 
 	rrr := &btpb.ReadRowsResponse{}
-	for col, cells := range r.cells {
-		i := strings.Index(col, ":") // guaranteed to exist
-		fam, col := col[:i], col[i+1:]
-		if len(cells) == 0 {
-			continue
-		}
-		// TODO(dsymonds): Apply transformers.
-		for _, cell := range cells {
-			rrr.Chunks = append(rrr.Chunks, &btpb.ReadRowsResponse_CellChunk{
-				RowKey:          []byte(r.key),
-				FamilyName:      &wrappers.StringValue{Value: fam},
-				Qualifier:       &wrappers.BytesValue{Value: []byte(col)},
-				TimestampMicros: cell.ts,
-				Value:           cell.value,
-			})
+	families := r.sortedFamilies()
+	for _, fam := range families {
+		for _, colName := range fam.colNames {
+			cells := fam.cells[colName]
+			if len(cells) == 0 {
+				continue
+			}
+			// TODO(dsymonds): Apply transformers.
+			for _, cell := range cells {
+				rrr.Chunks = append(rrr.Chunks, &btpb.ReadRowsResponse_CellChunk{
+					RowKey:          []byte(r.key),
+					FamilyName:      &wrappers.StringValue{Value: fam.name},
+					Qualifier:       &wrappers.BytesValue{Value: []byte(colName)},
+					TimestampMicros: cell.ts,
+					Value:           cell.value,
+				})
+			}
 		}
 	}
 	// We can't have a cell with just COMMIT set, which would imply a new empty cell.
@@ -388,7 +399,9 @@ func filterRow(f *btpb.RowFilter, r *row) bool {
 	switch f := f.Filter.(type) {
 	case *btpb.RowFilter_Chain_:
 		for _, sub := range f.Chain.Filters {
-			filterRow(sub, r)
+			if !filterRow(sub, r) {
+				return false
+			}
 		}
 		return true
 	case *btpb.RowFilter_Interleave_:
@@ -400,24 +413,42 @@ func filterRow(f *btpb.RowFilter, r *row) bool {
 		}
 		// merge
 		// TODO(dsymonds): is this correct?
-		r.cells = make(map[string][]cell)
+		r.families = make(map[string]*family)
 		for _, sr := range srs {
-			for col, cs := range sr.cells {
-				r.cells[col] = append(r.cells[col], cs...)
+			for _, fam := range sr.families {
+				f := r.getOrCreateFamily(fam.name, fam.order)
+				for colName, cs := range fam.cells {
+					f.cells[colName] = append(f.cellsByColumn(colName), cs...)
+				}
 			}
 		}
-		for _, cs := range r.cells {
-			sort.Sort(byDescTS(cs))
+		for _, fam := range r.families {
+			for _, cs := range fam.cells {
+				sort.Sort(byDescTS(cs))
+			}
 		}
 		return true
 	case *btpb.RowFilter_CellsPerColumnLimitFilter:
 		lim := int(f.CellsPerColumnLimitFilter)
-		for col, cs := range r.cells {
-			if len(cs) > lim {
-				r.cells[col] = cs[:lim]
+		for _, fam := range r.families {
+			for col, cs := range fam.cells {
+				if len(cs) > lim {
+					fam.cells[col] = cs[:lim]
+				}
 			}
 		}
 		return true
+	case *btpb.RowFilter_Condition_:
+		if filterRow(f.Condition.PredicateFilter, r.copy()) {
+			if f.Condition.TrueFilter == nil {
+				return false
+			}
+			return filterRow(f.Condition.TrueFilter, r)
+		}
+		if f.Condition.FalseFilter == nil {
+			return false
+		}
+		return filterRow(f.Condition.FalseFilter, r)
 	case *btpb.RowFilter_RowKeyRegexFilter:
 		pat := string(f.RowKeyRegexFilter)
 		rx, err := regexp.Compile(pat)
@@ -432,11 +463,11 @@ func filterRow(f *btpb.RowFilter, r *row) bool {
 
 	// Any other case, operate on a per-cell basis.
 	cellCount := 0
-	for key, cs := range r.cells {
-		i := strings.Index(key, ":") // guaranteed to exist
-		fam, col := key[:i], key[i+1:]
-		r.cells[key] = filterCells(f, fam, col, cs)
-		cellCount += len(r.cells[key])
+	for _, fam := range r.families {
+		for colName, cs := range fam.cells {
+			fam.cells[colName] = filterCells(f, fam.name, colName, cs)
+			cellCount += len(fam.cells[colName])
+		}
 	}
 	return cellCount > 0
 }
@@ -445,12 +476,26 @@ func filterCells(f *btpb.RowFilter, fam, col string, cs []cell) []cell {
 	var ret []cell
 	for _, cell := range cs {
 		if includeCell(f, fam, col, cell) {
+			cell = modifyCell(f, cell)
 			ret = append(ret, cell)
 		}
 	}
 	return ret
 }
 
+func modifyCell(f *btpb.RowFilter, c cell) cell {
+	if f == nil {
+		return c
+	}
+	// Consider filters that may modify the cell contents
+	switch f.Filter.(type) {
+	case *btpb.RowFilter_StripValueTransformer:
+		return cell{ts: c.ts}
+	default:
+		return c
+	}
+}
+
 func includeCell(f *btpb.RowFilter, fam, col string, cell cell) bool {
 	if f == nil {
 		return true
@@ -463,6 +508,9 @@ func includeCell(f *btpb.RowFilter, fam, col string, cell cell) bool {
 	case *btpb.RowFilter_RowKeyRegexFilter:
 		// Don't log, row-level filter
 		return true
+	case *btpb.RowFilter_StripValueTransformer:
+		// Don't log, cell-modifying filter
+		return true
 	default:
 		log.Printf("WARNING: don't know how to handle filter of type %T (ignoring it)", f)
 		return true
@@ -515,6 +563,25 @@ func includeCell(f *btpb.RowFilter, fam, col string, cell cell) bool {
 		// Lower bound is inclusive and defaults to 0, upper bound is exclusive and defaults to infinity.
 		return cell.ts >= f.TimestampRangeFilter.StartTimestampMicros &&
 			(f.TimestampRangeFilter.EndTimestampMicros == 0 || cell.ts < f.TimestampRangeFilter.EndTimestampMicros)
+	case *btpb.RowFilter_ValueRangeFilter:
+		v := cell.value
+		// Start value defaults to empty string closed
+		inRangeStart := func() bool { return bytes.Compare(v, []byte{}) >= 0 }
+		switch sv := f.ValueRangeFilter.StartValue.(type) {
+		case *btpb.ValueRange_StartValueOpen:
+			inRangeStart = func() bool { return bytes.Compare(v, sv.StartValueOpen) > 0 }
+		case *btpb.ValueRange_StartValueClosed:
+			inRangeStart = func() bool { return bytes.Compare(v, sv.StartValueClosed) >= 0 }
+		}
+		// End value defaults to no upper boundary
+		inRangeEnd := func() bool { return true }
+		switch ev := f.ValueRangeFilter.EndValue.(type) {
+		case *btpb.ValueRange_EndValueClosed:
+			inRangeEnd = func() bool { return bytes.Compare(v, ev.EndValueClosed) <= 0 }
+		case *btpb.ValueRange_EndValueOpen:
+			inRangeEnd = func() bool { return bytes.Compare(v, ev.EndValueOpen) < 0 }
+		}
+		return inRangeStart() && inRangeEnd()
 	}
 }
 
@@ -526,11 +593,11 @@ func (s *server) MutateRow(ctx context.Context, req *btpb.MutateRowRequest) (*bt
 		return nil, grpc.Errorf(codes.NotFound, "table %q not found", req.TableName)
 	}
 
-	fs := tbl.columnFamiliesSet()
+	fs := tbl.columnFamilies()
 	r := tbl.mutableRow(string(req.RowKey))
 	r.mu.Lock()
+	defer tbl.resortRowIndex() // Make sure the row lock is released before this grabs the table lock
 	defer r.mu.Unlock()
-
 	if err := applyMutations(tbl, r, req.Mutations, fs); err != nil {
 		return nil, err
 	}
@@ -547,8 +614,9 @@ func (s *server) MutateRows(req *btpb.MutateRowsRequest, stream btpb.Bigtable_Mu
 
 	res := &btpb.MutateRowsResponse{Entries: make([]*btpb.MutateRowsResponse_Entry, len(req.Entries))}
 
-	fs := tbl.columnFamiliesSet()
+	fs := tbl.columnFamilies()
 
+	defer tbl.resortRowIndex()
 	for i, entry := range req.Entries {
 		r := tbl.mutableRow(string(entry.RowKey))
 		r.mu.Lock()
@@ -577,7 +645,7 @@ func (s *server) CheckAndMutateRow(ctx context.Context, req *btpb.CheckAndMutate
 
 	res := &btpb.CheckAndMutateRowResponse{}
 
-	fs := tbl.columnFamiliesSet()
+	fs := tbl.columnFamilies()
 
 	r := tbl.mutableRow(string(req.RowKey))
 	r.mu.Lock()
@@ -587,18 +655,13 @@ func (s *server) CheckAndMutateRow(ctx context.Context, req *btpb.CheckAndMutate
 	whichMut := false
 	if req.PredicateFilter == nil {
 		// Use true_mutations iff row contains any cells.
-		whichMut = len(r.cells) > 0
+		whichMut = !r.isEmpty()
 	} else {
 		// Use true_mutations iff any cells in the row match the filter.
 		// TODO(dsymonds): This could be cheaper.
 		nr := r.copy()
 		filterRow(req.PredicateFilter, nr)
-		for _, cs := range nr.cells {
-			if len(cs) > 0 {
-				whichMut = true
-				break
-			}
-		}
+		whichMut = !nr.isEmpty()
 		// TODO(dsymonds): Figure out if this is supposed to be set
 		// even when there's no predicate filter.
 		res.PredicateMatched = whichMut
@@ -608,6 +671,7 @@ func (s *server) CheckAndMutateRow(ctx context.Context, req *btpb.CheckAndMutate
 		muts = req.TrueMutations
 	}
 
+	defer tbl.resortRowIndex()
 	if err := applyMutations(tbl, r, muts, fs); err != nil {
 		return nil, err
 	}
@@ -617,14 +681,14 @@ func (s *server) CheckAndMutateRow(ctx context.Context, req *btpb.CheckAndMutate
 // applyMutations applies a sequence of mutations to a row.
 // fam should be a snapshot of the keys of tbl.families.
 // It assumes r.mu is locked.
-func applyMutations(tbl *table, r *row, muts []*btpb.Mutation, fs map[string]bool) error {
+func applyMutations(tbl *table, r *row, muts []*btpb.Mutation, fs map[string]*columnFamily) error {
 	for _, mut := range muts {
 		switch mut := mut.Mutation.(type) {
 		default:
 			return fmt.Errorf("can't handle mutation type %T", mut)
 		case *btpb.Mutation_SetCell_:
 			set := mut.SetCell
-			if !fs[set.FamilyName] {
+			if _, ok := fs[set.FamilyName]; !ok {
 				return fmt.Errorf("unknown family %q", set.FamilyName)
 			}
 			ts := set.TimestampMicros
@@ -634,54 +698,65 @@ func applyMutations(tbl *table, r *row, muts []*btpb.Mutation, fs map[string]boo
 			if !tbl.validTimestamp(ts) {
 				return fmt.Errorf("invalid timestamp %d", ts)
 			}
-			col := fmt.Sprintf("%s:%s", set.FamilyName, set.ColumnQualifier)
+			fam := set.FamilyName
+			col := string(set.ColumnQualifier)
 
 			newCell := cell{ts: ts, value: set.Value}
-			r.cells[col] = appendOrReplaceCell(r.cells[col], newCell)
+			f := r.getOrCreateFamily(fam, fs[fam].order)
+			f.cells[col] = appendOrReplaceCell(f.cellsByColumn(col), newCell)
 		case *btpb.Mutation_DeleteFromColumn_:
 			del := mut.DeleteFromColumn
-			col := fmt.Sprintf("%s:%s", del.FamilyName, del.ColumnQualifier)
-
-			cs := r.cells[col]
-			if del.TimeRange != nil {
-				tsr := del.TimeRange
-				if !tbl.validTimestamp(tsr.StartTimestampMicros) {
-					return fmt.Errorf("invalid timestamp %d", tsr.StartTimestampMicros)
-				}
-				if !tbl.validTimestamp(tsr.EndTimestampMicros) {
-					return fmt.Errorf("invalid timestamp %d", tsr.EndTimestampMicros)
-				}
-				// Find half-open interval to remove.
-				// Cells are in descending timestamp order,
-				// so the predicates to sort.Search are inverted.
-				si, ei := 0, len(cs)
-				if tsr.StartTimestampMicros > 0 {
-					ei = sort.Search(len(cs), func(i int) bool { return cs[i].ts < tsr.StartTimestampMicros })
-				}
-				if tsr.EndTimestampMicros > 0 {
-					si = sort.Search(len(cs), func(i int) bool { return cs[i].ts < tsr.EndTimestampMicros })
-				}
-				if si < ei {
-					copy(cs[si:], cs[ei:])
-					cs = cs[:len(cs)-(ei-si)]
-				}
-			} else {
-				cs = nil
+			if _, ok := fs[del.FamilyName]; !ok {
+				return fmt.Errorf("unknown family %q", del.FamilyName)
 			}
-			if len(cs) == 0 {
-				delete(r.cells, col)
-			} else {
-				r.cells[col] = cs
+			fam := del.FamilyName
+			col := string(del.ColumnQualifier)
+			if _, ok := r.families[fam]; ok {
+				cs := r.families[fam].cells[col]
+				if del.TimeRange != nil {
+					tsr := del.TimeRange
+					if !tbl.validTimestamp(tsr.StartTimestampMicros) {
+						return fmt.Errorf("invalid timestamp %d", tsr.StartTimestampMicros)
+					}
+					if !tbl.validTimestamp(tsr.EndTimestampMicros) {
+						return fmt.Errorf("invalid timestamp %d", tsr.EndTimestampMicros)
+					}
+					// Find half-open interval to remove.
+					// Cells are in descending timestamp order,
+					// so the predicates to sort.Search are inverted.
+					si, ei := 0, len(cs)
+					if tsr.StartTimestampMicros > 0 {
+						ei = sort.Search(len(cs), func(i int) bool { return cs[i].ts < tsr.StartTimestampMicros })
+					}
+					if tsr.EndTimestampMicros > 0 {
+						si = sort.Search(len(cs), func(i int) bool { return cs[i].ts < tsr.EndTimestampMicros })
+					}
+					if si < ei {
+						copy(cs[si:], cs[ei:])
+						cs = cs[:len(cs)-(ei-si)]
+					}
+				} else {
+					cs = nil
+				}
+				if len(cs) == 0 {
+					delete(r.families[fam].cells, col)
+					colNames := r.families[fam].colNames
+					i := sort.Search(len(colNames), func(i int) bool { return colNames[i] >= col })
+					if i < len(colNames) && colNames[i] == col {
+						r.families[fam].colNames = append(colNames[:i], colNames[i+1:]...)
+					}
+					if len(r.families[fam].cells) == 0 {
+						delete(r.families, fam)
+					}
+				} else {
+					r.families[fam].cells[col] = cs
+				}
 			}
 		case *btpb.Mutation_DeleteFromRow_:
-			r.cells = make(map[string][]cell)
+			r.families = make(map[string]*family)
 		case *btpb.Mutation_DeleteFromFamily_:
-			fampre := mut.DeleteFromFamily.FamilyName + ":"
-			for col, _ := range r.cells {
-				if strings.HasPrefix(col, fampre) {
-					delete(r.cells, col)
-				}
-			}
+			fampre := mut.DeleteFromFamily.FamilyName
+			delete(r.families, fampre)
 		}
 	}
 	return nil
@@ -726,7 +801,7 @@ func (s *server) ReadModifyWriteRow(ctx context.Context, req *btpb.ReadModifyWri
 
 	updates := make(map[string]cell) // copy of updated cells; keyed by full column name
 
-	fs := tbl.columnFamiliesSet()
+	fs := tbl.columnFamilies()
 
 	r := tbl.mutableRow(string(req.RowKey))
 	r.mu.Lock()
@@ -734,17 +809,21 @@ func (s *server) ReadModifyWriteRow(ctx context.Context, req *btpb.ReadModifyWri
 	// Assume all mutations apply to the most recent version of the cell.
 	// TODO(dsymonds): Verify this assumption and document it in the proto.
 	for _, rule := range req.Rules {
-		if !fs[rule.FamilyName] {
+		if _, ok := fs[rule.FamilyName]; !ok {
 			return nil, fmt.Errorf("unknown family %q", rule.FamilyName)
 		}
 
-		key := fmt.Sprintf("%s:%s", rule.FamilyName, rule.ColumnQualifier)
+		fam := rule.FamilyName
+		col := string(rule.ColumnQualifier)
+		isEmpty := false
+		f := r.getOrCreateFamily(fam, fs[fam].order)
+		cs := f.cells[col]
+		isEmpty = len(cs) == 0
 
-		cells := r.cells[key]
 		ts := newTimestamp()
 		var newCell, prevCell cell
-		isEmpty := len(cells) == 0
 		if !isEmpty {
+			cells := r.families[fam].cells[col]
 			prevCell = cells[0]
 
 			// ts is the max of now or the prev cell's timestamp in case the
@@ -771,8 +850,9 @@ func (s *server) ReadModifyWriteRow(ctx context.Context, req *btpb.ReadModifyWri
 			binary.BigEndian.PutUint64(val[:], uint64(v))
 			newCell = cell{ts: ts, value: val[:]}
 		}
+		key := strings.Join([]string{fam, col}, ":")
 		updates[key] = newCell
-		r.cells[key] = appendOrReplaceCell(r.cells[key], newCell)
+		f.cells[col] = appendOrReplaceCell(f.cellsByColumn(col), newCell)
 	}
 
 	res := &btpb.Row{
@@ -872,6 +952,7 @@ func (s *server) gcloop(done <-chan int) {
 
 type table struct {
 	mu       sync.RWMutex
+	counter  uint64                   // increment by 1 when a new family is created
 	families map[string]*columnFamily // keyed by plain family name
 	rows     []*row                   // sorted by row key
 	rowIndex map[string]*row          // indexed by row key
@@ -879,16 +960,20 @@ type table struct {
 
 func newTable(ctr *btapb.CreateTableRequest) *table {
 	fams := make(map[string]*columnFamily)
+	c := uint64(0)
 	if ctr.Table != nil {
 		for id, cf := range ctr.Table.ColumnFamilies {
 			fams[id] = &columnFamily{
 				name:   ctr.Parent + "/columnFamilies/" + id,
+				order:  c,
 				gcRule: cf.GcRule,
 			}
+			c++
 		}
 	}
 	return &table{
 		families: fams,
+		counter:  c,
 		rowIndex: make(map[string]*row),
 	}
 }
@@ -908,14 +993,6 @@ func (t *table) columnFamilies() map[string]*columnFamily {
 	return cp
 }
 
-func (t *table) columnFamiliesSet() map[string]bool {
-	fs := make(map[string]bool)
-	for fam := range t.columnFamilies() {
-		fs[fam] = true
-	}
-	return fs
-}
-
 func (t *table) mutableRow(row string) *row {
 	// Try fast path first.
 	t.mu.RLock()
@@ -932,12 +1009,17 @@ func (t *table) mutableRow(row string) *row {
 		r = newRow(row)
 		t.rowIndex[row] = r
 		t.rows = append(t.rows, r)
-		sort.Sort(byRowKey(t.rows)) // yay, inefficient!
 	}
 	t.mu.Unlock()
 	return r
 }
 
+func (t *table) resortRowIndex() {
+	t.mu.Lock()
+	sort.Sort(byRowKey(t.rows))
+	t.mu.Unlock()
+}
+
 func (t *table) gc() {
 	// This method doesn't add or remove rows, so we only need a read lock for the table.
 	t.mu.RLock()
@@ -970,14 +1052,14 @@ func (b byRowKey) Less(i, j int) bool { return b[i].key < b[j].key }
 type row struct {
 	key string
 
-	mu    sync.Mutex
-	cells map[string][]cell // keyed by full column name; cells are in descending timestamp order
+	mu       sync.Mutex
+	families map[string]*family // keyed by family name
 }
 
 func newRow(key string) *row {
 	return &row{
-		key:   key,
-		cells: make(map[string][]cell),
+		key:      key,
+		families: make(map[string]*family),
 	}
 }
 
@@ -985,36 +1067,78 @@ func newRow(key string) *row {
 // Cell values are aliased.
 // r.mu should be held.
 func (r *row) copy() *row {
-	nr := &row{
-		key:   r.key,
-		cells: make(map[string][]cell, len(r.cells)),
-	}
-	for col, cs := range r.cells {
-		// Copy the []cell slice, but not the []byte inside each cell.
-		nr.cells[col] = append([]cell(nil), cs...)
+	nr := newRow(r.key)
+	for _, fam := range r.families {
+		nr.families[fam.name] = &family{
+			name:     fam.name,
+			order:    fam.order,
+			colNames: fam.colNames,
+			cells:    make(map[string][]cell),
+		}
+		for col, cs := range fam.cells {
+			// Copy the []cell slice, but not the []byte inside each cell.
+			nr.families[fam.name].cells[col] = append([]cell(nil), cs...)
+		}
 	}
 	return nr
 }
 
+// isEmpty returns true if a row doesn't contain any cell
+func (r *row) isEmpty() bool {
+	for _, fam := range r.families {
+		for _, cs := range fam.cells {
+			if len(cs) > 0 {
+				return false
+			}
+		}
+	}
+	return true
+}
+
+// sortedFamilies returns a column family set
+// sorted in ascending creation order in a row.
+func (r *row) sortedFamilies() []*family {
+	var families []*family
+	for _, fam := range r.families {
+		families = append(families, fam)
+	}
+	sort.Sort(byCreationOrder(families))
+	return families
+}
+
+func (r *row) getOrCreateFamily(name string, order uint64) *family {
+	if _, ok := r.families[name]; !ok {
+		r.families[name] = &family{
+			name:  name,
+			order: order,
+			cells: make(map[string][]cell),
+		}
+	}
+	return r.families[name]
+}
+
 // gc applies the given GC rules to the row.
 // r.mu should be held.
 func (r *row) gc(rules map[string]*btapb.GcRule) {
-	for col, cs := range r.cells {
-		fam := col[:strings.Index(col, ":")]
-		rule, ok := rules[fam]
+	for _, fam := range r.families {
+		rule, ok := rules[fam.name]
 		if !ok {
 			continue
 		}
-		r.cells[col] = applyGC(cs, rule)
+		for col, cs := range fam.cells {
+			r.families[fam.name].cells[col] = applyGC(cs, rule)
+		}
 	}
 }
 
 // size returns the total size of all cell values in the row.
 func (r *row) size() int {
 	size := 0
-	for _, cells := range r.cells {
-		for _, cell := range cells {
-			size += len(cell.value)
+	for _, fam := range r.families {
+		for _, cells := range fam.cells {
+			for _, cell := range cells {
+				size += len(cell.value)
+			}
 		}
 	}
 	return size
@@ -1061,6 +1185,29 @@ func applyGC(cells []cell, rule *btapb.GcRule) []cell {
 	return cells
 }
 
+type family struct {
+	name     string            // Column family name
+	order    uint64            // Creation order of column family
+	colNames []string          // Collumn names are sorted in lexicographical ascending order
+	cells    map[string][]cell // Keyed by collumn name; cells are in descending timestamp order
+}
+
+type byCreationOrder []*family
+
+func (b byCreationOrder) Len() int           { return len(b) }
+func (b byCreationOrder) Swap(i, j int)      { b[i], b[j] = b[j], b[i] }
+func (b byCreationOrder) Less(i, j int) bool { return b[i].order < b[j].order }
+
+// cellsByColumn adds the column name to colNames set if it does not exist
+// and returns all cells within a column
+func (f *family) cellsByColumn(name string) []cell {
+	if _, ok := f.cells[name]; !ok {
+		f.colNames = append(f.colNames, name)
+		sort.Strings(f.colNames)
+	}
+	return f.cells[name]
+}
+
 type cell struct {
 	ts    int64
 	value []byte
@@ -1074,6 +1221,7 @@ func (b byDescTS) Less(i, j int) bool { return b[i].ts > b[j].ts }
 
 type columnFamily struct {
 	name   string
+	order  uint64 // Creation order of column family
 	gcRule *btapb.GcRule
 }
 
