@@ -35,26 +35,42 @@ import (
 	"time"
 
 	"cloud.google.com/go/bigtable"
-	"cloud.google.com/go/bigtable/internal/cbtrc"
+	"cloud.google.com/go/bigtable/internal/cbtconfig"
 	"golang.org/x/net/context"
 	"google.golang.org/api/option"
+	"google.golang.org/grpc"
 )
 
 var (
 	oFlag = flag.String("o", "", "if set, redirect stdout to this file")
 
-	config              *cbtrc.Config
+	config              *cbtconfig.Config
 	client              *bigtable.Client
 	adminClient         *bigtable.AdminClient
 	instanceAdminClient *bigtable.InstanceAdminClient
+
+	version      = "<unknown version>"
+	revision     = "<unknown revision>"
+	revisionDate = "<unknown revision date>"
 )
 
+func getCredentialOpts(opts []option.ClientOption) []option.ClientOption {
+	if ts := config.TokenSource; ts != nil {
+		opts = append(opts, option.WithTokenSource(ts))
+	}
+	if tlsCreds := config.TLSCreds; tlsCreds != nil {
+		opts = append(opts, option.WithGRPCDialOption(grpc.WithTransportCredentials(tlsCreds)))
+	}
+	return opts
+}
+
 func getClient() *bigtable.Client {
 	if client == nil {
 		var opts []option.ClientOption
 		if ep := config.DataEndpoint; ep != "" {
 			opts = append(opts, option.WithEndpoint(ep))
 		}
+		opts = getCredentialOpts(opts)
 		var err error
 		client, err = bigtable.NewClient(context.Background(), config.Project, config.Instance, opts...)
 		if err != nil {
@@ -70,6 +86,7 @@ func getAdminClient() *bigtable.AdminClient {
 		if ep := config.AdminEndpoint; ep != "" {
 			opts = append(opts, option.WithEndpoint(ep))
 		}
+		opts = getCredentialOpts(opts)
 		var err error
 		adminClient, err = bigtable.NewAdminClient(context.Background(), config.Project, config.Instance, opts...)
 		if err != nil {
@@ -85,6 +102,7 @@ func getInstanceAdminClient() *bigtable.InstanceAdminClient {
 		if ep := config.AdminEndpoint; ep != "" {
 			opts = append(opts, option.WithEndpoint(ep))
 		}
+		opts = getCredentialOpts(opts)
 		var err error
 		instanceAdminClient, err = bigtable.NewInstanceAdminClient(context.Background(), config.Project, opts...)
 		if err != nil {
@@ -96,7 +114,7 @@ func getInstanceAdminClient() *bigtable.InstanceAdminClient {
 
 func main() {
 	var err error
-	config, err = cbtrc.Load()
+	config, err = cbtconfig.Load()
 	if err != nil {
 		log.Fatal(err)
 	}
@@ -104,9 +122,6 @@ func main() {
 
 	flag.Usage = func() { usage(os.Stderr) }
 	flag.Parse()
-	if config.Creds != "" {
-		os.Setenv("GOOGLE_APPLICATION_CREDENTIALS", config.Creds)
-	}
 	if flag.NArg() == 0 {
 		usage(os.Stderr)
 		os.Exit(1)
@@ -161,90 +176,94 @@ func init() {
 var configHelp = `
 For convenience, values of the -project, -instance, -creds,
 -admin-endpoint and -data-endpoint flags may be specified in
-` + cbtrc.Filename() + ` in this format:
+` + cbtconfig.Filename() + ` in this format:
 	project = my-project-123
 	instance = my-instance
 	creds = path-to-account-key.json
 	admin-endpoint = hostname:port
 	data-endpoint = hostname:port
 All values are optional, and all will be overridden by flags.
+
+cbt ` + version + ` ` + revision + ` ` + revisionDate + `
 `
 
 var commands = []struct {
 	Name, Desc string
 	do         func(context.Context, ...string)
 	Usage      string
-	Required   cbtrc.RequiredFlags
+	Required   cbtconfig.RequiredFlags
 }{
 	{
 		Name:     "count",
 		Desc:     "Count rows in a table",
 		do:       doCount,
 		Usage:    "cbt count <table>",
-		Required: cbtrc.ProjectAndInstanceRequired,
+		Required: cbtconfig.ProjectAndInstanceRequired,
 	},
 	{
 		Name:     "createfamily",
 		Desc:     "Create a column family",
 		do:       doCreateFamily,
 		Usage:    "cbt createfamily <table> <family>",
-		Required: cbtrc.ProjectAndInstanceRequired,
+		Required: cbtconfig.ProjectAndInstanceRequired,
 	},
 	{
-		Name:     "createtable",
-		Desc:     "Create a table",
-		do:       doCreateTable,
-		Usage:    "cbt createtable <table>",
-		Required: cbtrc.ProjectAndInstanceRequired,
+		Name: "createtable",
+		Desc: "Create a table",
+		do:   doCreateTable,
+		Usage: "cbt createtable <table> [initial_splits...]\n" +
+			"  initial_splits=row		A row key to be used to initially split the table " +
+			"into multiple tablets. Can be repeated to create multiple splits.",
+		Required: cbtconfig.ProjectAndInstanceRequired,
 	},
 	{
 		Name:     "deletefamily",
 		Desc:     "Delete a column family",
 		do:       doDeleteFamily,
 		Usage:    "cbt deletefamily <table> <family>",
-		Required: cbtrc.ProjectAndInstanceRequired,
+		Required: cbtconfig.ProjectAndInstanceRequired,
 	},
 	{
 		Name:     "deleterow",
 		Desc:     "Delete a row",
 		do:       doDeleteRow,
 		Usage:    "cbt deleterow <table> <row>",
-		Required: cbtrc.ProjectAndInstanceRequired,
+		Required: cbtconfig.ProjectAndInstanceRequired,
 	},
 	{
 		Name:     "deletetable",
 		Desc:     "Delete a table",
 		do:       doDeleteTable,
 		Usage:    "cbt deletetable <table>",
-		Required: cbtrc.ProjectAndInstanceRequired,
+		Required: cbtconfig.ProjectAndInstanceRequired,
 	},
 	{
 		Name:     "doc",
 		Desc:     "Print godoc-suitable documentation for cbt",
 		do:       doDoc,
 		Usage:    "cbt doc",
-		Required: cbtrc.NoneRequired,
+		Required: cbtconfig.NoneRequired,
 	},
 	{
 		Name:     "help",
 		Desc:     "Print help text",
 		do:       doHelp,
 		Usage:    "cbt help [command]",
-		Required: cbtrc.NoneRequired,
+		Required: cbtconfig.NoneRequired,
 	},
 	{
 		Name:     "listinstances",
 		Desc:     "List instances in a project",
 		do:       doListInstances,
 		Usage:    "cbt listinstances",
-		Required: cbtrc.ProjectRequired,
+		Required: cbtconfig.ProjectRequired,
 	},
 	{
 		Name:     "lookup",
 		Desc:     "Read from a single row",
 		do:       doLookup,
 		Usage:    "cbt lookup <table> <row>",
-		Required: cbtrc.ProjectAndInstanceRequired,
+		Required: cbtconfig.ProjectAndInstanceRequired,
 	},
 	{
 		Name: "ls",
@@ -252,14 +271,14 @@ var commands = []struct {
 		do:   doLS,
 		Usage: "cbt ls			List tables\n" +
 			"cbt ls <table>		List column families in <table>",
-		Required: cbtrc.ProjectAndInstanceRequired,
+		Required: cbtconfig.ProjectAndInstanceRequired,
 	},
 	{
 		Name:     "mddoc",
 		Desc:     "Print documentation for cbt in Markdown format",
 		do:       doMDDoc,
 		Usage:    "cbt mddoc",
-		Required: cbtrc.NoneRequired,
+		Required: cbtconfig.NoneRequired,
 	},
 	{
 		Name: "read",
@@ -270,7 +289,7 @@ var commands = []struct {
 			"  end=<row>		Stop reading before this row\n" +
 			"  prefix=<prefix>	Read rows with this prefix\n" +
 			"  count=<n>		Read only this many rows\n",
-		Required: cbtrc.ProjectAndInstanceRequired,
+		Required: cbtconfig.ProjectAndInstanceRequired,
 	},
 	{
 		Name: "set",
@@ -282,7 +301,7 @@ var commands = []struct {
 			"  ts is an optional integer timestamp.\n" +
 			"  If it cannot be parsed, the `@ts` part will be\n" +
 			"  interpreted as part of the value.",
-		Required: cbtrc.ProjectAndInstanceRequired,
+		Required: cbtconfig.ProjectAndInstanceRequired,
 	},
 	{
 		Name: "setgcpolicy",
@@ -292,7 +311,14 @@ var commands = []struct {
 			"\n" +
 			`  maxage=<d>		Maximum timestamp age to preserve (e.g. "1h", "4d")` + "\n" +
 			"  maxversions=<n>	Maximum number of versions to preserve",
-		Required: cbtrc.ProjectAndInstanceRequired,
+		Required: cbtconfig.ProjectAndInstanceRequired,
+	},
+	{
+		Name:     "version",
+		Desc:     "Print the current cbt version",
+		do:       doVersion,
+		Usage:    "cbt version",
+		Required: cbtconfig.NoneRequired,
 	},
 }
 
@@ -324,10 +350,16 @@ func doCreateFamily(ctx context.Context, args ...string) {
 }
 
 func doCreateTable(ctx context.Context, args ...string) {
-	if len(args) != 1 {
-		log.Fatal("usage: cbt createtable <table>")
+	if len(args) < 1 {
+		log.Fatal("usage: cbt createtable <table> [initial_splits...]")
+	}
+	var err error
+	if len(args) > 1 {
+		splits := args[1:]
+		err = getAdminClient().CreatePresplitTable(ctx, args[0], splits)
+	} else {
+		err = getAdminClient().CreateTable(ctx, args[0])
 	}
-	err := getAdminClient().CreateTable(ctx, args[0])
 	if err != nil {
 		log.Fatalf("Creating table: %v", err)
 	}
@@ -542,6 +574,12 @@ func (b byColumn) Len() int           { return len(b) }
 func (b byColumn) Swap(i, j int)      { b[i], b[j] = b[j], b[i] }
 func (b byColumn) Less(i, j int) bool { return b[i].Column < b[j].Column }
 
+type byFamilyName []bigtable.FamilyInfo
+
+func (b byFamilyName) Len() int           { return len(b) }
+func (b byFamilyName) Swap(i, j int)      { b[i], b[j] = b[j], b[i] }
+func (b byFamilyName) Less(i, j int) bool { return b[i].Name < b[j].Name }
+
 func doLS(ctx context.Context, args ...string) {
 	switch len(args) {
 	default:
@@ -561,10 +599,14 @@ func doLS(ctx context.Context, args ...string) {
 		if err != nil {
 			log.Fatalf("Getting table info: %v", err)
 		}
-		sort.Strings(ti.Families)
-		for _, fam := range ti.Families {
-			fmt.Println(fam)
+		sort.Sort(byFamilyName(ti.FamilyInfos))
+		tw := tabwriter.NewWriter(os.Stdout, 10, 8, 4, '\t', 0)
+		fmt.Fprintf(tw, "Family Name\tGC Policy\n")
+		fmt.Fprintf(tw, "-----------\t---------\n")
+		for _, fam := range ti.FamilyInfos {
+			fmt.Fprintf(tw, "%s\t%s\n", fam.Name, fam.GCPolicy)
 		}
+		tw.Flush()
 	}
 }
 
@@ -768,3 +810,7 @@ var unitMap = map[string]time.Duration{
 	"h":  time.Hour,
 	"d":  24 * time.Hour,
 }
+
+func doVersion(ctx context.Context, args ...string) {
+	fmt.Printf("%s %s %s\n", version, revision, revisionDate)
+}
