@@ -21,6 +21,8 @@ import (
 	"os"
 	"reflect"
 
+	"cloud.google.com/go/internal/version"
+
 	"github.com/golang/protobuf/proto"
 	"golang.org/x/net/context"
 	"google.golang.org/api/option"
@@ -51,39 +53,45 @@ type protoClient interface {
 // datastoreClient is a wrapper for the pb.DatastoreClient that includes gRPC
 // metadata to be sent in each request for server-side traffic management.
 type datastoreClient struct {
+	// Embed so we still implement the DatastoreClient interface,
+	// if the interface adds more methods.
+	pb.DatastoreClient
+
 	c  pb.DatastoreClient
 	md metadata.MD
 }
 
 func newDatastoreClient(conn *grpc.ClientConn, projectID string) pb.DatastoreClient {
 	return &datastoreClient{
-		c:  pb.NewDatastoreClient(conn),
-		md: metadata.Pairs(resourcePrefixHeader, "projects/"+projectID),
+		c: pb.NewDatastoreClient(conn),
+		md: metadata.Pairs(
+			resourcePrefixHeader, "projects/"+projectID,
+			"x-goog-api-client", fmt.Sprintf("gl-go/%s gccl/%s grpc/", version.Go(), version.Repo)),
 	}
 }
 
 func (dc *datastoreClient) Lookup(ctx context.Context, in *pb.LookupRequest, opts ...grpc.CallOption) (*pb.LookupResponse, error) {
-	return dc.c.Lookup(metadata.NewContext(ctx, dc.md), in, opts...)
+	return dc.c.Lookup(metadata.NewOutgoingContext(ctx, dc.md), in, opts...)
 }
 
 func (dc *datastoreClient) RunQuery(ctx context.Context, in *pb.RunQueryRequest, opts ...grpc.CallOption) (*pb.RunQueryResponse, error) {
-	return dc.c.RunQuery(metadata.NewContext(ctx, dc.md), in, opts...)
+	return dc.c.RunQuery(metadata.NewOutgoingContext(ctx, dc.md), in, opts...)
 }
 
 func (dc *datastoreClient) BeginTransaction(ctx context.Context, in *pb.BeginTransactionRequest, opts ...grpc.CallOption) (*pb.BeginTransactionResponse, error) {
-	return dc.c.BeginTransaction(metadata.NewContext(ctx, dc.md), in, opts...)
+	return dc.c.BeginTransaction(metadata.NewOutgoingContext(ctx, dc.md), in, opts...)
 }
 
 func (dc *datastoreClient) Commit(ctx context.Context, in *pb.CommitRequest, opts ...grpc.CallOption) (*pb.CommitResponse, error) {
-	return dc.c.Commit(metadata.NewContext(ctx, dc.md), in, opts...)
+	return dc.c.Commit(metadata.NewOutgoingContext(ctx, dc.md), in, opts...)
 }
 
 func (dc *datastoreClient) Rollback(ctx context.Context, in *pb.RollbackRequest, opts ...grpc.CallOption) (*pb.RollbackResponse, error) {
-	return dc.c.Rollback(metadata.NewContext(ctx, dc.md), in, opts...)
+	return dc.c.Rollback(metadata.NewOutgoingContext(ctx, dc.md), in, opts...)
 }
 
 func (dc *datastoreClient) AllocateIds(ctx context.Context, in *pb.AllocateIdsRequest, opts ...grpc.CallOption) (*pb.AllocateIdsResponse, error) {
-	return dc.c.AllocateIds(metadata.NewContext(ctx, dc.md), in, opts...)
+	return dc.c.AllocateIds(metadata.NewOutgoingContext(ctx, dc.md), in, opts...)
 }
 
 // Client is a client for reading and writing data in a datastore dataset.
@@ -198,9 +206,9 @@ func keyToProto(k *Key) *pb.Key {
 	for {
 		el := &pb.Key_PathElement{Kind: k.Kind}
 		if k.ID != 0 {
-			el.IdType = &pb.Key_PathElement_Id{k.ID}
+			el.IdType = &pb.Key_PathElement_Id{Id: k.ID}
 		} else if k.Name != "" {
-			el.IdType = &pb.Key_PathElement_Name{k.Name}
+			el.IdType = &pb.Key_PathElement_Name{Name: k.Name}
 		}
 		path = append([]*pb.Key_PathElement{el}, path...)
 		if k.Parent == nil {
@@ -407,14 +415,29 @@ func (c *Client) get(ctx context.Context, keys []*Key, dst interface{}, opts *pb
 	if err != nil {
 		return err
 	}
-	if len(resp.Deferred) > 0 {
-		// TODO(jbd): Assess whether we should retry the deferred keys.
-		return errors.New("datastore: some entities temporarily unavailable")
+	found := resp.Found
+	missing := resp.Missing
+	// Upper bound 100 iterations to prevent infinite loop.
+	// We choose 100 iterations somewhat logically:
+	// Max number of Entities you can request from Datastore is 1,000.
+	// Max size for a Datastore Entity is 1 MiB.
+	// Max request size is 10 MiB, so we assume max response size is also 10 MiB.
+	// 1,000 / 10 = 100.
+	// Note that if ctx has a deadline, the deadline will probably
+	// be hit before we reach 100 iterations.
+	for i := 0; len(resp.Deferred) > 0 && i < 100; i++ {
+		req.Keys = resp.Deferred
+		resp, err = c.client.Lookup(ctx, req)
+		if err != nil {
+			return err
+		}
+		found = append(found, resp.Found...)
+		missing = append(missing, resp.Missing...)
 	}
-	if len(keys) != len(resp.Found)+len(resp.Missing) {
+	if len(keys) != len(found)+len(missing) {
 		return errors.New("datastore: internal error: server returned the wrong number of entities")
 	}
-	for _, e := range resp.Found {
+	for _, e := range found {
 		k, err := protoToKey(e.Entity.Key)
 		if err != nil {
 			return errors.New("datastore: internal error: server returned an invalid key")
@@ -427,12 +450,12 @@ func (c *Client) get(ctx context.Context, keys []*Key, dst interface{}, opts *pb
 		if multiArgType == multiArgTypeStructPtr && elem.IsNil() {
 			elem.Set(reflect.New(elem.Type().Elem()))
 		}
-		if err := loadEntity(elem.Interface(), e.Entity); err != nil {
+		if err := loadEntityProto(elem.Interface(), e.Entity); err != nil {
 			multiErr[index] = err
 			any = true
 		}
 	}
-	for _, e := range resp.Missing {
+	for _, e := range missing {
 		k, err := protoToKey(e.Entity.Key)
 		if err != nil {
 			return errors.New("datastore: internal error: server returned an invalid key")
@@ -530,9 +553,9 @@ func putMutations(keys []*Key, src interface{}) ([]*pb.Mutation, error) {
 		}
 		var mut *pb.Mutation
 		if k.Incomplete() {
-			mut = &pb.Mutation{Operation: &pb.Mutation_Insert{p}}
+			mut = &pb.Mutation{Operation: &pb.Mutation_Insert{Insert: p}}
 		} else {
-			mut = &pb.Mutation{Operation: &pb.Mutation_Upsert{p}}
+			mut = &pb.Mutation{Operation: &pb.Mutation_Upsert{Upsert: p}}
 		}
 		mutations = append(mutations, mut)
 	}
@@ -574,7 +597,7 @@ func deleteMutations(keys []*Key) ([]*pb.Mutation, error) {
 			return nil, fmt.Errorf("datastore: can't delete the incomplete key: %v", k)
 		}
 		mutations = append(mutations, &pb.Mutation{
-			Operation: &pb.Mutation_Delete{keyToProto(k)},
+			Operation: &pb.Mutation_Delete{Delete: keyToProto(k)},
 		})
 	}
 	return mutations, nil
